import datetime
import functools
import io
import typing
from unittest.mock import MagicMock

from botocore.client import ClientMeta
from botocore.config import Config

import lobotomy
from lobotomy import _services
from lobotomy._clients import _casting
from lobotomy._clients import _validation


class Client:
    """
    Mock AWS boto3.client behaviors.

    This pulls response data from data stored in configuration files.
    """

    def __init__(
        self,
        session: "lobotomy.Session",
        service_name: str,
        region_name: str | None = None,
        api_version: str | None = None,
        use_ssl: bool = True,
        verify: None | bool | str = None,
        endpoint_url: str | None = None,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        config: Config | None = None,
    ):
        """Populate with the loaded scenario data."""
        self._service_name = service_name
        self._client_init_kwargs: dict = {
            "region_name": region_name,
            "api_version": api_version,
            "use_ssl": use_ssl,
            "verify": verify,
            "endpoint_url": endpoint_url,
            "aws_access_key_id": aws_access_key_id,
            "aws_secret_access_key": aws_secret_access_key,
            "aws_session_token": aws_session_token,
            "config": config,
        }
        self._calls: list[lobotomy.ServiceCall] = []
        self._service: _services.Service = _services.load_definition(
            service_name,
        )
        self._session = session

        self.meta = ClientMeta(
            events=MagicMock(),
            client_config=config,
            endpoint_url=endpoint_url,
            service_model=MagicMock(),
            method_to_api_mapping={},
            partition="aws",
        )
        self.exceptions = MagicMock(ClientError=lobotomy.ClientError)
        self._registered_exceptions = {}
        for name in self._service.exceptions:
            new_type = type(name, (lobotomy.ClientError,), {})
            setattr(self.exceptions, name, new_type)
            self._registered_exceptions[name] = new_type

    def _call(self, called_method_name: str, *args, **kwargs):
        raw = self._session.lobotomy.pop_response(
            self._service_name,
            called_method_name,
            arguments={"args": args, "kwargs": kwargs},
        )
        method = self._service.lookup(called_method_name)
        request = _validation.validate_input(method, args, kwargs)
        response = _casting.cast(method.output, raw)

        # The s3.download_fileobj is a special case where the data is written to the
        # Fileobj argument and not returned via an output argument.
        if self._service_name == "s3" and called_method_name == "download_fileobj":
            buffer: io.BytesIO = request["Fileobj"]
            file_obj_data: str | bytes | None = response.pop("Fileobj", None)
            if isinstance(file_obj_data, str):
                file_obj_data = file_obj_data.encode()
            if file_obj_data:
                buffer.write(file_obj_data)

        service_call = lobotomy.ServiceCall(
            service=self._service_name,
            method=called_method_name,
            request=request,
            args=args,
            kwargs=kwargs,
            response=response,
        )

        self._session.lobotomy.record_call(service_call)
        self._calls.append(service_call)

        if "Error" in response:
            error = self._registered_exceptions.get(
                response["Error"]["Code"],
                lobotomy.ClientError,
            )()
            raise error.populate(
                operation_name=(
                    called_method_name.replace("_", " ").title().replace(" ", "")
                ),
                **response["Error"],
            )
        return response

    def __getattr__(self, item: str):
        """
        Retrieve response data for the given client call.

        This comes from the scenario data that defines the execution.
        """
        if not self._service.has(item):
            raise lobotomy.NoSuchMethodError(
                f"""
                No boto/botocore definition found for "{self._service_name}{item}()".
                The version of botocore installed could be out of date, or newer
                and this service method has been removed. Please check the boto3
                documentation to confirm the existence of this service method.
                """
            )

        return functools.partial(self._call, item)

    def generate_presigned_url(
        self,
        ClientMethod: str,  # noqa: N803
        Params: dict[str, typing.Any] | None = None,  # noqa: N803
        ExpiresIn: int = 3600,  # noqa: N803
        HttpMethod: str | None = None,  # noqa: N803
    ):
        """
        Generate a presigned url given a client, its method, and arguments.

        https://docs.aws.amazon.com/STS/latest/APIReference/CommonParameters.html

        :param ClientMethod:
            The client method to presign for.
        :param Params:
            The parameters normally passed to ClientMethod.
        :param ExpiresIn:
            The number of seconds the presigned url is valid for. By default it
            expires in an hour (3600 seconds)
        :param HttpMethod:
            The http method to use on the generated url. By default, the http method
            is whatever is used in the method's model.
        """
        name = self._service_name
        action = "".join([w.capitalize() for w in ClientMethod.split("_")])
        version = self._client_init_kwargs["api_version"] or "2011-06-15"
        access = (
            self._client_init_kwargs["aws_access_key_id"]
            or self._session.get_credentials().access_key
        )
        region = self._client_init_kwargs["region_name"] or self._session.region_name
        now = datetime.datetime.now(datetime.UTC).replace(microsecond=0)
        day = now.date().isoformat().replace("-", "")
        date = (
            now.isoformat("T").replace("+00:00", "Z").replace(":", "").replace("-", "")
        )
        url_parts = [
            f"https://{name}.amazonaws.com/",
            f"?Action={action}",
            f"&Version={version}",
            "&X-Amz-Algorithm=AWS4-HMAC-SHA256",
            f"&X-Amz-Credential={access}%2F{day}%2F{region}%2F{name}%2Faws4_request",
            f"&X-Amz-Date={date}",
            f"&X-Amz-Expires={ExpiresIn}",
            "&X-Amz-SignedHeaders=host",
            "&X-Amz-Security-Token=LobotomyFakeToken",
            "&X-Amz-Signature=lobotomyfakesignature",
        ]
        return "".join(url_parts)

    def get_paginator(self, item: str) -> MagicMock:
        """Mock a single-page paginator response."""

        def _call(*args, **kwargs):
            return [self._call(item, *args, **kwargs)]

        paginator = MagicMock()
        paginator.paginate.side_effect = _call
        return paginator



import dataclasses
import json
import pathlib
import re
import textwrap
import typing

import toml
import yaml
import yaml.constructor

_indent_regex = re.compile(r"^(?P<indent>\s*)")
_path_regex = re.compile(r"!lobotomy.inject_(?P<kind>[^\s]+)\s+(?P<path>[^\s\n]+)")


@dataclasses.dataclass(frozen=True)
class YamlBlock:
    """Data structure for extracted YAML blocks."""

    #: Key within the larger context of the YAML data that this block data is
    #: associated with. The root YAML block should have an empty string for a key.
    key: str
    #: Line that precedes the body that contains the key.
    key_line: str
    #: The string contents of for the given key, not including the line with the
    #: key itself on it.
    body: str
    #: Absolute index where the block body starts in the YAML file that it was
    #: extracted from.
    start_index: int
    #: Absolute index where the block body ends (exclusive) in the YAML file
    #: that it was extracted from.
    end_index: int

    @property
    def key_index(self) -> int:
        """
        Get index of the line where the key resides.

        This will be one less than the index where the body starts, unless the body
        starts at 0 because this is the root block.
        """
        return max(0, self.start_index - 1)

    @property
    def outer_body(self) -> str:
        """
        Fetch the body and the key line together.

        This pulls both instead of just the body. which contains only
        the inner contents of the key.
        """
        if not self.key:
            return self.body

        return "{}\n{}".format(self.key_line, textwrap.indent(self.body, "  "))


def _get_block(key: str, block: YamlBlock) -> YamlBlock:
    """
    Retrieve the block within the specified argument block's body for the specified key.

    The body of a block has no key header line as the body is the string
    representation of the contents within the given block. Therefore, they search
    key, if present in the block, will be a top-level key in the block.

    :param key:
        Search key to extract from the specified block.
    :param block:
        Parent block in which to search for the specified key.
    :return:
        A new block containing the body contents of the specified key extracted from
        the contents of the parent block. If the key is not found a block will be
        returned with an empty body and start/end indexes of -1.
    """
    lines = block.body.split("\n")
    index_finder = (
        i for i, line in enumerate(lines) if line.strip().startswith(f"{key}:")
    )
    start = next(index_finder, None)
    if start is None:
        return YamlBlock(key, f"{key}:", "", -1, -1)

    output_lines = []
    for line in lines[(start + 1) :]:
        if not line.startswith(" "):
            break
        output_lines.append(line)

    offset = block.start_index + start + 1
    return YamlBlock(
        key,
        key_line=lines[start],
        body=textwrap.dedent("\n".join(output_lines)),
        start_index=offset,
        end_index=offset + len(output_lines),
    )


def _extract(
    contents: str,
    prefix: typing.Iterable[str],
) -> tuple[YamlBlock, YamlBlock, YamlBlock]:
    """
    Extract the lobotomy sessions and clients blocks from the overall yaml block.

    This returns them along with the lobotomy key-prefix yaml block in which the clients
    and sessions blocks were found.

    :param contents:
        YAML string contents from which to extract lobotomy data.
    :param prefix:
        Key prefix within the YAML string where the lobotomy data resides.
    :return:
        Clients block, sessions block, and parent lobotomy block.
    """
    body = textwrap.dedent(contents.strip().replace("\r", ""))
    prefix_block = YamlBlock("", "", body, 0, 1 + body.count("\n"))
    for key in prefix:
        prefix_block = _get_block(key, prefix_block)

    return (
        _get_block("clients", prefix_block),
        _get_block("sessions", prefix_block),
        prefix_block,
    )


def _get_prefix(
    prefix: None | str | typing.Iterable[str],
) -> typing.Iterable[str]:
    """
    Fetch the optional prefix keys defining where the lobotomy data will be nested.

    Optional prefix that specifies where within the data dictionary to find the
    lobotomized data. If not specified, the data dictionary is assumed to be a
    lobotomy data configuration dictionary. If specified, the prefix can either
    be represented as a hierarchical list of keys to descend within the root
    data dictionary, e.g. ["foo", "bar"] or as a dot-delimited string denoting
    the hierarchical list of keys, e.g. "foo.bar". If one of the keys contains
    a "." then the list form must be used.
    """
    if not prefix:
        return []

    return prefix.split(".") if isinstance(prefix, str) else prefix


def _normalize_yaml(contents: str, directory: pathlib.Path) -> str:
    """
    Normalize Yaml class configuration for advanced loading functionality.

    :param contents:
        String contents of a YAML file to be parsed.
    :param directory:
        Absolute directory to the folder location for the loaded YAML file.
    """

    def path_replacer(match: re.Match) -> str:
        kind = match.group("kind")
        source = match.group("path").strip("\"'")
        p = directory.joinpath(source).expanduser().absolute()
        value = json.dumps({"absolute": str(p), "original": source})
        return f"!lobotomy.inject_{kind} '{value}'"

    normalized = re.sub(_path_regex, path_replacer, contents)
    return normalized


def _read_yaml(
    contents: str,
    prefix: typing.Iterable[str],
    directory: pathlib.Path,
) -> dict:
    """
    Read a YAML file into lobotomy data.

    First it tries to read the entire file, but if that does not work it falls back to
    extracting the lobotomy data from the file and then loading only that portion of
    the file. This is useful when YAML classes are used within the file that may not be
    loaded at the time of the file read process.

    :param contents:
        YAML string contents from which to extract lobotomy data.
    :param prefix:
        Key prefix within the YAML string where the lobotomy data resides.
    :param directory:
        Directory to the folder in which the YAML file resides.
    :return:
        Lobotomy data loaded from the YAML string by either of the methods.
    """
    normalized = _normalize_yaml(contents, directory)
    try:
        output = yaml.full_load(normalized)
        for key in prefix:
            output = output.get(key) or {}
        return {k: v for k, v in output.items() if k in ("clients", "sessions")}
    except yaml.constructor.ConstructorError:
        pass

    clients_block, sessions_block, _ = _extract(normalized, prefix)
    data = {}
    if clients_block.start_index != -1:
        data.update(yaml.full_load(clients_block.outer_body))
    if sessions_block.start_index != -1:
        data.update(yaml.full_load(sessions_block.outer_body))
    return data


def _read_file_data(
    path: pathlib.Path,
    file_format: str | None,
) -> dict:
    """Read existing data from the source file with the given format."""
    contents = path.read_text()

    if file_format == "yaml" or path.name.endswith((".yaml", ".yml")):
        return yaml.full_load(_normalize_yaml(contents, path.parent))

    if file_format == "toml" or path.name.endswith(".toml"):
        return typing.cast(dict, toml.loads(contents))

    return json.loads(contents)


def _read_file(
    path: pathlib.Path,
    file_format: str | None,
    prefix: typing.Iterable[str],
) -> dict:
    """
    Read the configuration file from disk if it exists.

    :param path:
        Location of the file to read.
    :param file_format:
        Specifies the format of the file to load. If None the extension of the
        file path will be used to determine the type of file to read.
    :param prefix:
        Key prefix within the YAML string where the lobotomy data resides.
    :return:
        Lobotomy data loaded from the file at the specified prefix.
    """
    p = path.expanduser().absolute()
    if not p.exists():
        raise FileNotFoundError(f"Missing file {p}")

    try:
        output = _read_file_data(p, file_format) or {}
    except yaml.constructor.ConstructorError:
        return _read_yaml(p.read_text(), prefix, p.parent)

    for key in prefix:
        output = output.get(key) or {}
    return {k: v for k, v in output.items() if k in ("clients", "sessions")}


def read(
    path: str | pathlib.Path,
    prefix: str | typing.Iterable[str] | None = None,
    file_format: str | None = None,
) -> dict[str, typing.Any]:
    """
    Read the specified configuration file and returns the result.

    If a prefix is specified, the data at the given prefix will be returned instead.

    :param path:
        Path to the configuration file to read.
    :param prefix:
        Optional prefix within the file in which the configuration resides.
    :param file_format:
        Optional format for the file, which will be determined from the
        file extension if omitted.
    :return:
        A dictionary containing the lobotomy configuration data from the
        given file. This will be an empty dictionary if no configuration
        exists yet.
    """
    return _read_file(pathlib.Path(path), file_format, _get_prefix(prefix))


def _update_yaml_write_lines(
    lines: list[str],
    configs: dict[str, typing.Any],
    block: YamlBlock,
    parent: YamlBlock,
) -> list[str]:
    """
    Replace the specified block within the YAML file lines object with the new configs.

    This is done by dumping that configs body and replacing it within the returned list
    of lines.

    :param lines:
        Complete YAML file lines for the file that will be written. A modified
        version of these will be returned that remove the existing lines for
        the target block and replace them with the new data stored in the lobotomy
        data configs object. If the new configuration is empty for the block, the
        existing lines will be extracted and no new lines added to the returned
        lines list.
    :param configs:
        Lobotomy data from which the new block configuration will be generated.
    :param block:
        YamlBlock to be updated in the output with the new configs data.
    :param parent:
        Parent YamlBlock in which the specified block resides, which is used to
        handle cases where the block definition is new.
    :return:
        A modified version of the lines argument where the old block data has
        been removed and replaced by the new block data serialized from the
        specified lobotomy data configs.
    """
    if parent.key == "":
        # If there is no key for the parent, it's the root of the file and there
        # will be no indent.
        indent = ""
    else:
        index = parent.key_index if block.start_index == -1 else block.key_index
        indent = _indent_regex.match(lines[index]).group("indent")  # type: ignore

    if data := configs.get(block.key):
        new_body = [textwrap.indent(yaml.dump({block.key: data}), indent).rstrip()]
    else:
        # Don't include the body if the configuration is empty.
        new_body = []

    if block.start_index == -1:
        before = lines[: (parent.key_index + 1)]
        after = lines[(parent.key_index + 1) :]
    else:
        before = lines[: block.key_index]
        after = lines[block.end_index :]
    return before + new_body + after


def _write_modified_yaml(
    path: pathlib.Path,
    configs: dict[str, typing.Any],
    prefix: typing.Iterable[str],
) -> None:
    """
    Write the updated lobotomy data configs to an existing YAML file.

    This is done without loading the entire file to prevent issues with YAML classes.

    :param path:
        Path where the lobotomy data will be written.
    :param configs:
        Lobotomy data to write to the file.
    :param prefix:
        Key hierarchy where the lobotomy data will be written within the file.
    """
    body = textwrap.dedent(path.read_text().strip().replace("\r", ""))
    lines = body.split("\n")
    clients_block, sessions_block, parent_block = _extract(body, prefix)

    # Sort the blocks so that they are modified from bottom of the file to
    # the top so that line index changes don't corrupt earlier block index
    # values.
    blocks = sorted(
        [clients_block, sessions_block],
        key=lambda b: b.start_index,
        reverse=True,
    )
    for block in blocks:
        lines = _update_yaml_write_lines(lines, configs, block, parent_block)

    body = "{}\n".format("\n".join(lines))
    path.write_text(body)


def _get_data_for_write(
    path: pathlib.Path,
    file_format: str,
    prefix_keys: typing.Iterable[str],
) -> tuple[dict[str, typing.Any], dict[str, typing.Any]]:
    """
    Retrieve the data from the file if it exists or create a new data object if not.

    :param path:
        Location of the source file to load data from if it already exists.
    :param file_format:
        Specifies the format of the file to read.
    :param prefix_keys:
        A list of keys within the root data object where the lobotomy data should be
        written. If the iterable is empty, the lobotomy data will be written at the
        root level.
    :return:
        A tuple containing the root data object and the parent data object for the
        lobotomy write insertions. If there are no prefix keys specified, the root
        and the parent will be the same. Otherwise, the parent will be a nested child
        of the root at the location specified by the prefix keys.
    """
    root = _read_file_data(path, file_format) if path.exists() else {}
    parent = root
    for key in prefix_keys or []:
        if key not in parent:
            parent[key] = {}
        parent = parent[key]

    return root, parent


def _get_file_format_id(path: pathlib.Path, file_format: str | None) -> str:
    """Determine the file format for writing based on the arguments."""
    formats = {
        "yaml": path.name.endswith((".yml", ".yaml")),
        "toml": path.name.endswith(".toml"),
        "json": path.name.endswith(".json"),
    }
    finder = (k for k, v in formats.items() if file_format == k or v)
    return next(finder, "json")


def write(
    path: str | pathlib.Path,
    configs: dict[str, typing.Any],
    prefix: None | str | typing.Iterable[str] = None,
    file_format: str | None = None,
) -> None:
    """
    Write the specified configuration file with updated configuration data.

    If a prefix is specified, the data will be inserted at the given prefix.

    :param path:
        Path to the configuration file to read.
    :param configs:
        Updated lobotomy configuration data to be written to the file.
        This will replace the existing configuration data in the file.
    :param prefix:
        Optional prefix within the file in which the configuration resides.
    :param file_format:
        Optional format for the file, which will be determined from the
        file extension if omitted.
    :return:
        A dictionary containing the lobotomy configuration data from the
        given file. This will be an empty dictionary if no configuration
        exists yet.
    """
    prefix_keys = _get_prefix(prefix)
    p = pathlib.Path(path).expanduser().absolute()

    # Make sure the configs are clean of everything but the expected lobotomy keys.
    source = {k: v for k, v in configs.items() if k in ("clients", "sessions")}

    format_id = _get_file_format_id(p, file_format)

    if p.exists() and format_id == "yaml":
        _write_modified_yaml(p, source, prefix_keys)
        return

    data, parent = _get_data_for_write(p, format_id, prefix_keys)
    parent.update(source)

    writers = {
        "yaml": lambda: yaml.dump(data),
        "toml": lambda: toml.dumps(data),
        "json": lambda: json.dumps(data, indent=2),
    }

    contents = writers[format_id]()
    p.write_text(contents)



from lobotomy._cli import _adder
from lobotomy._cli import _definitions
from lobotomy._cli import _parsing


def run(
    arguments: list[str] | None = None,
) -> "_definitions.ExecutionResult":
    """
    Execute a cli command with the given arguments.

    Defaults to sys.argv if no explicit arguments were specified.
    """
    args = _parsing.parse(arguments)
    context = _definitions.CliContext(args)

    commands = {
        "add": _adder.run,
    }

    return commands[args.command](context)



import datetime
import typing
from unittest.mock import MagicMock

import dateutil.parser
from botocore.response import StreamingBody

import lobotomy


class InternalEventStreamer:
    """
    Mock representation of an EventStream object.

    This is used notably in the S3 Select return body. The source class for this
    is located at `botocore.eventstream.EventStream`.
    """

    def __init__(
        self,
        source: (
            dict[str, typing.Any]
            | list[dict[str, typing.Any]]
            | tuple[dict[str, typing.Any], ...]
        ),
    ):
        """Wrap the response in a streaming interface."""
        self._source = [source] if isinstance(source, dict) else source

    def __iter__(self):
        """Iterate over each element in the source."""
        yield from self._source


class InternalStreamer:
    """
    Mock representation of the StreamingBody object.

    An internal class that replaces the HTTPResponse object that boto
    StreamingBody objects wrap. This is used to simulate the streaming
    body behaviors in returned streaming response objects.
    """

    def __init__(self, source: bytes | str):
        """Wrap the source string in a streaming interface."""
        self._cursor = 0
        self._source: bytes | str = source

        # This exists for botocore compatibility setting timeouts.
        self._fp = MagicMock()

    def read(self, amt: int | None = None):
        """Simulate the streaming interface."""
        length = len(self._source)
        amount = amt if amt is not None else length
        start = max(0, self._cursor)
        end = min(length, self._cursor + amount)
        self._cursor = end
        return self._source[start:end]

    def close(self):
        """
        Close the stream.

        Exists only for compatibility cases as this isn't used here.
        """
        pass

    def __len__(self):
        """Get length of the internal streaming body."""
        return len(self._source)


def _cast_structure(
    definition: dict,
    value: dict | list,
) -> dict | InternalEventStreamer:
    """
    Convert a dictionary/structure botocore type into its formatted values.

    This recursively casts all the members of the dictionary/structure.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        The cast version of the specified value that matches the format of
        the value as it would be returned in a boto client response.
    """
    if definition.get("eventstream"):
        sub_definition = definition.copy()
        sub_definition["eventstream"] = False
        values = [value] if isinstance(value, dict) else value
        return InternalEventStreamer([cast(sub_definition, v) for v in values])

    return {
        k: cast(definition["members"].get(k), v)
        for k, v in typing.cast(dict, value).items()
    }


def _cast_noop(definition: dict, value: typing.Any) -> typing.Any:
    """
    Cast as a no-op operation used for unknown or uncastable value types.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        An unmodified value as this is a no-op.
    """
    return value


def _cast_integer(definition: dict, value: typing.Any) -> int:
    """
    Convert a value into an integer value.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        The cast version of the specified value that matches the format of
        the value as it would be returned in a boto client response.
    """
    return int(value)


def _cast_list(definition: dict, value: list) -> list:
    """
    Convert a list botocore type into formatted values recursively casting its items.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        The cast version of the specified value that matches the format of
        the value as it would be returned in a boto client response.
    """
    return [cast(definition.get("member"), v) for v in value]


def _cast_string(
    definition: dict,
    value: typing.Any,
) -> str | StreamingBody:
    """
    Convert a string botocore type into its formatted value.

    If the definition indicates that the string is returned in streaming format,
    a StreamingBody object is returned instead with the string value
    injected into the body via an InternalStreamer object that mimics the
    actual streaming behavior of a real StreamingBody response.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        The cast version of the specified value that matches the format of
        the value as it would be returned in a boto client response.
    """
    output = str(value)
    if definition.get("streaming"):
        output = StreamingBody(InternalStreamer(output), len(output))
    return output


def _cast_blob(
    definition: dict,
    value: typing.Any,
) -> bytes | StreamingBody:
    """
    Convert a blob botocore type into its formatted value.

    If the definition indicates that the blob is returned in streaming format,
    a StreamingBody object is returned instead with the bytes value
    injected into the body via an InternalStreamer object that mimics the
    actual streaming behavior of a real StreamingBody response.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        The cast version of the specified value that matches the format of
        the value as it would be returned in a boto client response.
    """
    output = value if isinstance(value, bytes) else str(value).encode()

    if definition.get("streaming"):
        output = StreamingBody(InternalStreamer(output), len(output))
    return output


def _cast_timestamp(
    definition: dict,
    value: float | int | str | datetime.date | datetime.datetime,
) -> datetime.datetime:
    """
    Convert a string date definition into a boto-style datetime.

    This matches the format as would be returned by a boto response. The dateutil
    library is used to parse the string for flexible representation. Additionally, an
    integer/float value can be specified, which will be treated as a unix timestamp.
    Note that boto returns timezone-aware datetime objects, so this loading process
    ensures that the returned datetime is timezone-aware. If no timezone
    was determined by the loading process, UTC is used by default.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        The cast version of the specified value that matches the format of
        the value as it would be returned in a boto client response.
    """
    if isinstance(value, (int, float)):
        output = datetime.datetime.fromtimestamp(value, tz=datetime.UTC)
    elif isinstance(value, datetime.datetime):
        output = value
    elif isinstance(value, datetime.date):
        output = datetime.datetime(
            year=value.year,
            month=value.month,
            day=value.day,
            tzinfo=datetime.UTC,
        )
    else:
        output = dateutil.parser.parse(value)

    if not output.tzinfo:
        return output.replace(tzinfo=datetime.UTC)
    return output


def cast(definition: dict | None, value: typing.Any) -> typing.Any:
    """
    Cast the raw mocked value into its equivalent boto response type.

    Determines the casting from the data type as defined by the definition object for
    the value. For complex types, this function acts recursively.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        The cast version of the specified value that matches the format of
        the value as it would be returned in a boto client response.
    """
    if definition is None:
        # If there is no definition because it does not exist, just return
        # the value as-is.
        return value

    data_type: str | None = definition.get("type")
    conversions = {
        "structure": _cast_structure,
        "list": _cast_list,
        "timestamp": _cast_timestamp,
        "string": _cast_string,
        "long": _cast_integer,
        "integer": _cast_integer,
        "blob": _cast_blob,
        "noop": _cast_noop,
    }
    caster: typing.Any = conversions.get(
        data_type or "noop",
        _cast_noop,
    )

    try:
        if isinstance(value, lobotomy.YamlModifier):
            return caster(definition, value.to_response_data())
        return caster(definition, value)
    except Exception as error:
        raise lobotomy.DataTypeError(
            f"""
            Failed to cast specified response data of type "{data_type}"
            for the value:
            {value}
            There may be an incompatibility between the value specified in
            the lobotomy configuration data for the response and the data type
            expected by the return of the boto response.
            """
        ) from error



from lobotomy._services._definitions import Method  # noqa: F401
from lobotomy._services._definitions import Service  # noqa: F401


def load_definition(service_name: str) -> "Service":
    """Load client definition data for the associated AWS service."""
    return Service(service_name)



import typing

import botocore.exceptions


class LobotomyError(Exception):
    """Base exception class for all lobotomy custom exceptions."""

    def __init__(self, message: str):
        """Create the generic lobotomy error with the given message."""
        super().__init__(message)


class NoResponseFoundError(LobotomyError):
    """Error raised when no response was found for the given method call."""

    pass


class NoSuchMethodError(LobotomyError):
    """Error raised when a service method definition cannot be found."""

    pass


class DataTypeError(LobotomyError):
    """Error raised during casting of data types that fails."""

    pass


class RequestValidationError(LobotomyError):
    """Error raised when validation of the request arguments fails for a method call."""

    pass


class ErrorResponseError(typing.TypedDict):
    """Botocore error response error definition."""

    Code: str
    Message: str | None


class ErrorResponse(typing.TypedDict):
    """Botocore error response object interface."""

    Error: "ErrorResponseError"


class ClientError(botocore.exceptions.ClientError):
    """
    An exception that mirrors the structure of a boto3/botocore exception.

    This is used to create errors without the added complexity of those actual exception
    classes.
    """

    def __init__(
        self,
        error_response: typing.Optional["ErrorResponse"] = None,
        operation_name: str | None = None,
        **kwargs,
    ):
        """Create the error."""
        response = error_response or {
            "Error": {"Code": "Unknown", "Message": "Unknown"}
        }
        name = operation_name or "Unknown"
        super().__init__(response, name)

    def populate(self, **kwargs) -> "ClientError":
        """Specify code and message for the client error."""
        code = kwargs.get("Code", "Unknown")
        message = kwargs.get("Message", "Unknown")
        error = {"Code": code, "Message": message}
        self.response = {"Error": error}
        self.operation_name = kwargs.get("operation_name", "Unknown")
        exception_message = self.MSG_TEMPLATE.format(
            error_code=code,
            error_message=message,
            operation_name=self.operation_name,
            retry_info="",
        )
        self.args = (exception_message,)
        return self



import copy
import datetime
import typing


def parse_definition_item(shapes: dict, value: dict, max_depth: int = 10) -> dict:
    """
    Unwrap the value object by placing its shape data within a copy of the object.

    Return that copied object. This works recursively to unwrap results hierarchically.

    :param shapes:
        Shape definitions from the service spec in which this value
        definition resides.
    :param value:
        A definition object value to unwrap shape data within, or potentially
        a value object that has no shape and is already a leaf definition
        value as this function is called recursively.
    :param max_depth:
        Max recursion of the parsing before the process aborts. This prevents circular
        reference issues in cases like dynamoDB table definitions where the definitions
        map back and forth.
    :return:
        An unwrapped, deep-copied version of the value dictionary where
        the shape(s) have been injected into the object hierarchically to
        remove the need for shape lookups to understand the properties,
        including types, of the value and potentially its member(s) if
        represents a list or dictionary.
    """
    output = copy.deepcopy(value)
    depth = max_depth - 1
    if depth < 0:
        # In cases where recursion is a problem, force an exit and set the
        # deep-value type to "any" to allow for any kind of validation.
        if "type" not in output:
            output["type"] = "any"
        return output

    if value.get("type") == "map":
        # Maps are terminal but have key and value shapes.
        output["key"] = parse_definition_item(shapes, value["key"], depth)
        output["value"] = parse_definition_item(shapes, value["value"], depth)

    if "member" in output:
        # Types like lists have a member definition for the items in the
        # list. That needs to be parsed and then the shape is complete.
        output["member"] = parse_definition_item(shapes, output["member"], depth)

    if "members" in output:
        output["members"] = {
            k: parse_definition_item(shapes, v, depth)
            for k, v in (output["members"] or {}).items()
        }

    if "shape" in output:
        shape = copy.deepcopy(shapes[output["shape"]])
        output.update(parse_definition_item(shapes, shape, depth))

    return output


def flat_cast(output: dict) -> typing.Any:
    """
    Create a flat version of the output.

    Default will be included for inclusion as a skeleton in a configuration
    response file.
    """
    o = output
    data_type = o.get("type", "structure")
    type_casts = {
        "blob": lambda: "...",
        "string": lambda: "...",
        "integer": lambda: 1,
        "long": lambda: 1,
        "float": lambda: 1.0,
        "timestamp": lambda: f"{datetime.datetime.utcnow().isoformat('T')}Z",
        "list": lambda: [flat_cast(o["member"])],
        # Some calls have empty requests or responses that have no members and so here
        # we allow for the empty possibility for the structure type.
        "structure": lambda: {k: flat_cast(v) for k, v in o.get("members", {}).items()},
    }
    caster = type_casts.get(data_type, lambda: None)
    return caster()



import dataclasses
import json
import pathlib
import typing

import yaml
import yaml.constructor


@dataclasses.dataclass()
class YamlModifier:
    """Base class for lobotomy YAML modifier class."""

    value: typing.Any

    def to_response_data(self) -> typing.Any:
        """Convert the modifier to its response format."""
        return None

    @classmethod
    def label(cls) -> str:
        """Get the yaml identifier representation for the class."""
        return "!lobotomy.unknown"

    @classmethod
    def _from_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "YamlModifier":
        """Load an internal yaml node parsing, defaulting to a scalar value."""
        value = loader.construct_scalar(typing.cast(yaml.ScalarNode, node))
        return cls(value)

    @classmethod
    def parse_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "YamlModifier":
        """Parse yaml node into this class object for Lobotomy processing."""
        return cls._from_yaml(loader, node)

    @classmethod
    def _dump_yaml(cls, dumper: yaml.Dumper, source: "YamlModifier") -> typing.Any:
        """Convert to a yaml node representation for writing to file."""
        return dumper.represent_scalar(source.label(), source.value)

    @classmethod
    def dump_yaml(cls, dumper: yaml.Dumper, source: "YamlModifier") -> typing.Any:
        """Dump to a yaml node representation for writing to file."""
        return cls._dump_yaml(dumper, source)

    @classmethod
    def register(cls):
        """Register the comparator with the PyYaml loader."""
        yaml.add_constructor(cls.label(), cls.parse_yaml)
        yaml.add_representer(cls, cls.dump_yaml)


class ToJsonEncoder(json.JSONEncoder):
    """Custom encoder that handles nested ToJson structures."""

    def default(self, value: typing.Any) -> typing.Any:
        """Handle unknown values to the standard JSON encoder."""
        if isinstance(value, ToJson):
            return value.to_response_data()


@dataclasses.dataclass()
class ToJson(YamlModifier):
    """YAML class for converting a YAML object into a JSON string in a response."""

    def to_response_data(self) -> typing.Any:
        """Convert the modifier to its response format."""
        return json.dumps(self.value, cls=ToJsonEncoder)

    @classmethod
    def label(cls) -> str:
        """Get the yaml identifier representation for the class."""
        return "!lobotomy.to_json"

    @classmethod
    def _from_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "ToJson":
        """Load an internal yaml node parsing."""
        try:
            node = typing.cast(yaml.MappingNode, node)
            return cls(loader.construct_mapping(node, deep=True))
        except yaml.constructor.ConstructorError:
            node = typing.cast(yaml.SequenceNode, node)
            return cls(loader.construct_sequence(node, deep=True))

    @classmethod
    def _dump_yaml(cls, dumper: yaml.Dumper, source: "YamlModifier") -> typing.Any:
        """Convert to a yaml node representation for writing to file."""
        if isinstance(source.value, (list, tuple)):
            return dumper.represent_sequence(source.label(), source.value)
        return dumper.represent_mapping(source.label(), source.value)


@dataclasses.dataclass()
class InjectString(YamlModifier):
    """YAML class for including an external file as a string value."""

    def to_response_data(self) -> typing.Any:
        """Convert the modifier to its response format."""
        return pathlib.Path(self.value["absolute"]).expanduser().absolute().read_text()

    @classmethod
    def label(cls) -> str:
        """Get the yaml identifier representation for the class."""
        return "!lobotomy.inject_string"

    @classmethod
    def _from_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "InjectString":
        """Load an internal yaml node parsing."""
        raw = loader.construct_scalar(typing.cast(yaml.ScalarNode, node))
        value = json.loads(typing.cast(str, raw).strip("\"'"))
        return cls(value)

    @classmethod
    def _dump_yaml(cls, dumper: yaml.Dumper, source: "YamlModifier") -> typing.Any:
        """Convert to a yaml node representation for writing to file."""
        return dumper.represent_scalar(source.label(), source.value["original"])


@dataclasses.dataclass()
class BotoError(YamlModifier):
    """YAML class for specifying boto errors for service calls."""

    def to_response_data(self) -> typing.Any:
        """Convert the modifier to its response format."""
        v = self.value or {}
        error_code = v.get("code", "GenericLobotomyError")
        error_message = v.get("message", "There was an error.")
        return {"Error": {"Code": error_code, "Message": error_message}}

    @classmethod
    def label(cls) -> str:
        """Get the yaml identifier representation for the class."""
        return "!lobotomy.error"

    @classmethod
    def _from_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "BotoError":
        """Load an internal yaml node parsing."""
        node = typing.cast(yaml.MappingNode, node)
        value = loader.construct_mapping(node, deep=True)
        return cls(value)

    @classmethod
    def _dump_yaml(cls, dumper: yaml.Dumper, source: "YamlModifier") -> typing.Any:
        """Convert to a yaml node representation for writing to file."""
        return dumper.represent_mapping(source.label(), source.value or {})


ToJson.register()
InjectString.register()
BotoError.register()



import dataclasses
import gzip
import importlib.resources
import json
import pathlib
import typing

import botocore
import yaml

from lobotomy._services import _formatting


@dataclasses.dataclass(frozen=True)
class DataWrapper:
    """
    Data structure class containing a data object with nested lookup.

    The get method allows for hierarchical, optional-chaining lookup of keys
    within the data object.
    """

    data: dict

    def get(self, *args: str, default: typing.Any = None) -> typing.Any:
        """
        Fetch the value from the spec dictionary.

        This method allows for nested key lookups by specifying multiple key args. If
        the key is not found in the spec dictionary, the default value will be returned
        instead. However, if a falsy value is found for the key, including None, that
        will be returned instead.
        """
        value = self.data or {}
        for k in args:
            if k not in value:
                return default
            value = value[k]
        return value


@dataclasses.dataclass(frozen=True)
class Method(DataWrapper):
    """Client method definition data structure."""

    #: Method name for the service client represented by this object.
    name: str
    #: Method operation definition from the service specification.
    data: dict
    #: Service in which this method resides.
    service: "Service"

    @property
    def input(self) -> dict:
        """Fetch definition for the input signature of this method."""
        return _formatting.parse_definition_item(
            self.service.shapes,
            self.get("input") or {},
        )

    @property
    def output(self) -> dict:
        """Fetch definition for the output/response signature of this method."""
        return _formatting.parse_definition_item(
            self.service.shapes,
            self.get("output") or {},
        )

    @property
    def configuration_output(self) -> typing.Any:
        """
        Fetch a configuration representation of the method output.

        This is used in lobotomy service method configurations.
        """
        return _formatting.flat_cast(self.output)


@dataclasses.dataclass(frozen=True)
class Service(DataWrapper):
    """
    Botocore service definition for the specified service with accessors.

    Adds functionality for use in lobotomy mocking situations.
    """

    #: Name of the service for the associated specification.
    name: str
    #: Dictionary containing the loaded botocore specification for the
    #: given service name.
    data: dict = dataclasses.field(init=False, default_factory=lambda: {})
    #: The exceptions for the associated service.
    exceptions: dict = dataclasses.field(init=False, default_factory=lambda: {})

    def __post_init__(self):
        """Load the service specification into the object."""
        self.data.update(_get_specification(self.name))
        self.exceptions.update(_get_exceptions(self.data))

    @property
    def version(self) -> str:
        """Fetch the version of the service specification."""
        return self.get("version", "1.0")

    @property
    def metadata(self) -> dict:
        """
        Fetch metadata associated with the service.

        This data is defined in the botocore specification.
        """
        return self.get("metadata") or {}

    @property
    def operations(self) -> dict:
        """
        Fetch a dictionary that maps client methods to their specifications.

        The method keys are all lowercase format, so it's best to use
        the Service.lookup(method_name) to retrieve the values for
        snake_case definitions.
        """
        return self.get("operations") or {}

    @property
    def shapes(self) -> dict:
        """
        Fetch the shape definitions from the service specification.

        This defines the input, output and exception type objects and their arguments.
        """
        return self.get("shapes") or {}

    def has(self, method_name: str) -> bool:
        """Determine whether the spec defines the given method."""
        value = method_name.lower().replace("_", "")
        return value in self.operations

    def lookup(self, method_name: str) -> "Method":
        """Fetch the Method data for the associated method name."""
        value = method_name.lower().replace("_", "")
        return Method(
            name=method_name,
            data=self.operations.get(value) or {},
            service=self,
        )


def _read_potential_gzip(path: pathlib.Path) -> str:
    """
    Read the contents of the given path, which may be gzipped.

    :param path:
        Path to the file to read.
    :return:
        Contents of the file as a string.
    """
    if not path.exists():
        path = path.with_suffix(f"{path.suffix}.gz")
        return gzip.decompress(path.read_bytes()).decode("utf-8")
    return path.read_text(encoding="utf-8")


def _get_specification(service_name: str) -> dict:
    """
    Load the service specification data for the given service name.

    This data is loaded from the installed botocore library for maximum compatibility.

    :param service_name:
        Name of the service to load, which corresponds with the name that
        would be given to create the service client object.
    :return:
        Dictionary containing the loaded JSON service specification for the
        botocore service. The operations section is modified such that the
        keys (i.e. method names) are all lowercase because the service spec
        defines them as PascalCase and that is difficult to lookup when the
        snake_case names are the ones invoked on the client.
    """
    directory = (
        pathlib.Path(botocore.__file__).parent.joinpath("data", service_name).absolute()
    )
    folder = next(
        n for n in sorted(directory.iterdir(), reverse=True) if n.name.startswith("20")
    )
    spec = json.loads(
        # As of botocore 1.32.1, the service definitions are gzipped.
        # This checks if the non-gzipped version exists and if not, tries
        # the gzipped version.
        _read_potential_gzip(directory.joinpath(folder, "service-2.json"))
    )

    # Augment the botocore service definitions with client additions added by boto3
    # for non-standard api operations. The notable example here is s3.upload_file,
    # which is a convenience wrapper around the much lower-level multipart upload
    # api methods.
    package = f"{__package__}._augmentations"
    resource_name = f"{service_name}.yaml"
    if importlib.resources.is_resource(package, resource_name):
        extras = yaml.full_load(importlib.resources.read_text(package, resource_name))

        for key, value in (extras.get("operations") or {}).items():
            spec["operations"][key] = value

        for key, value in (extras.get("shapes") or {}).items():
            spec["shapes"][key] = value

    spec["operations"] = {k.lower(): v for k, v in spec["operations"].items()}
    return spec


def _get_exceptions(specification: dict) -> dict[str, dict]:
    """
    Assemble a mapping of client exceptions raised by the methods.

    These are extracted from the specified service specification where the values are
    the shape mappings for those exceptions as specified in the service specification.

    :param specification:
        Botocore service specification from which to extract exception
        information.
    :return:
        Dictionary of exception names and their corresponding shapes
        set on the client.exceptions object.
    """
    errors = {
        error["shape"]
        for item in specification["operations"].values()
        for error in (item.get("errors") or [])
    }
    return {e: specification["shapes"][e] for e in errors}



import typing

import lobotomy
from lobotomy import _services


def validate_input(
    method: "_services.Method",
    request_args: typing.Iterable[typing.Any],
    request_kwargs: dict[str, typing.Any],
) -> dict[str, typing.Any]:
    """
    Compare the request made to its botocore definition.

    This raises assertion errors if there are missing required arguments or if there
    are extraneous unknown arguments for the top-level request calls.

    Future work here would also look at required arguments hierarchically as
    well as use some patterns and mins/maxes specified by the botocore
    definition to confirm that the values given are also valid.

    :param method:
        Method definition object for the method that was called.
    :param request_args:
        Positional arguments made to the method call.
    :param request_kwargs:
        Keyword arguments made to the method call.
    :return:
        A dictionary containing the normalized request that combines args and kwargs
        into a single request kwargs dictionary.
    """
    definition = method.input
    keys = list(definition["members"].keys())
    request = {
        **{keys[i]: a for i, a in enumerate(request_args)},
        **(request_kwargs or {}),
    }

    specified_keys = request.keys()
    required = set(definition.get("required", []))

    if required > set(specified_keys):
        raise lobotomy.RequestValidationError(
            f"""
            Missing required arguments {required - set(specified_keys)}
            on {method.service.name}.{method.name}.
            """
        )

    unknown_keys = {
        k for k in (set(specified_keys) - set(keys)) if not k.startswith("_")
    }

    if unknown_keys:
        raise lobotomy.RequestValidationError(
            f"""
            Unknown arguments {unknown_keys} found on call to
            {method.service.name}.{method.name}.
            """
        )

    return request
